#include <Windows.h>
#include <sigscanner.h>
#include <string>
#include <vector>
#include <filesystem>
#include <algorithm>
#include <fstream>
#include <sstream>
#include <MinHook.h>
#include <map>

std::map<std::string, std::string> macros;
const std::string macroDir = "macros/";

const size_t lobbyMaxLength = 1024;
const size_t chatMaxLength = 256;
const size_t chatMaxMessages = 3;

void scanDir(std::string dir) {
	macros.clear();
	namespace fs = std::filesystem;
	fs::create_directories(dir);
	for (const auto &entry : fs::directory_iterator(dir)) {
		if (!entry.is_regular_file()) continue;
		auto filename = entry.path().filename().stem().string();
		auto extension = entry.path().extension().stem().string();
		std::transform(extension.begin(), extension.end(), extension.begin(), [](unsigned char c) { return std::tolower(c); });
		if (extension == ".txt") {
			std::ifstream in(entry);
			if (in.good()) {
				std::stringstream buffer;
				buffer << in.rdbuf();
				in.close();
				macros["/" + filename] = buffer.str();
			}
		}
	}
}

std::string listMacros() {
	scanDir(macroDir);
	std::stringstream ss;
	ss << "Available text macros:";
	for (auto &it : macros) {
		ss << "  " << it.first;
	}
	ss << "   (wkTextMacros-v1)";
	return ss.str();
}

std::vector<std::string> split(const std::string& s, char delimiter){
	std::vector<std::string> tokens;
	std::string token;
	std::istringstream tokenStream(s);
	while (std::getline(tokenStream, token, delimiter)) {
		tokens.push_back(token);
	}
	return tokens;
}

// Lobby

int (__stdcall *origLobbyDisplayMessage)(int a1, char *msg);
void showGreenText(void *This, std::string msg) {
	std::string buff = "SYS:wkTextMacros:ALL:" + msg;
	origLobbyDisplayMessage((int) This + 0x10318, (char *) buff.c_str());
}

void (__fastcall *addrLobbySendGreentext)(const char *msg, void *EDX, void *This, int a3, int a4);
char *addrMyNickname;
int (__fastcall *origLobbyClientCommands)(void *This, void *EDX, char **commandstrptr, char **argstrptr);


int __fastcall hookLobbyClientCommands(void *This, void *EDX, char **commandstrptr, char **argstrptr) {
	std::string command = "/" + std::string(commandstrptr[0]);
	std::string args = std::string(argstrptr[0]);

	if (command == "/macros") {
		showGreenText(This, listMacros());
		return 1;
	} else if (macros.find(command) != macros.end()) {
		std::stringstream ss;
		ss << addrMyNickname << " ";
		auto &msg = macros[command];
		ss << ((msg.length() > lobbyMaxLength) ? msg.substr(0, lobbyMaxLength) : msg);
		addrLobbySendGreentext(ss.str().c_str(), 0, This, 0, 0);
		return 1;
	}

	return origLobbyClientCommands(This, EDX, commandstrptr, argstrptr);
}

// In-game chat
int (__stdcall *origShowChatMessage)(DWORD a1, int color, const char *msg, int unk);
int (__stdcall *origConstructResources)(int a1);
DWORD addrResourceObject;

int callShowChatMessage(const char *msg, int color) {
	if (!addrResourceObject) return -1;
	return origShowChatMessage(addrResourceObject, color, msg, 1);
}

DWORD origOnChatInput;
int __stdcall callOriginalOnChatInput(int a1, const char *msg, int a3) {
	_asm mov ecx, a1
	_asm mov eax, msg
	_asm push a3
	_asm call origOnChatInput
}

int onChatInput(int a1, char *msg, int a3) {
	if(!strcmp(msg, "/macros")) {
		std::string macrolist = listMacros();
		if(macrolist.length() > chatMaxLength)
			macrolist = macrolist.substr(0, chatMaxLength);
		callShowChatMessage(macrolist.c_str(), 6);
		return 1;
	}
	else if(macros.find(msg) != macros.end()) {
		auto & macro = macros[msg];
		auto parts = split(macro, '\n');
		size_t i = 0;
		for(auto & part : parts) {
			std::string out = (part.length() > chatMaxLength) ? part.substr(0, chatMaxLength) : part;
			callOriginalOnChatInput(a1, out.c_str(), a3);
			i++;
			if(i >= chatMaxMessages) break;
		}
		return 1;
	}
	return 0;
}

int __stdcall hookConstructResources(int a1) {
	addrResourceObject = a1;
	return origConstructResources(a1);
}

#pragma optimize( "", off )
char *onchat_eax;
int onchat_ecx;
void __stdcall hookOnChatInput(int a3) {
	_asm mov onchat_eax, eax
	_asm mov onchat_ecx, ecx
	if (!onChatInput(onchat_ecx, (char *) onchat_eax, a3)) {
		callOriginalOnChatInput(onchat_ecx, onchat_eax, a3);
	}
}
#pragma optimize( "", on )


void hook(std::string name, DWORD pTarget, DWORD *pDetour, DWORD *ppOriginal) {
	if (!pTarget)
		throw std::runtime_error("Hook adress is null: " + name);
	if (MH_CreateHook((LPVOID) pTarget, (LPVOID) pDetour, (LPVOID *) ppOriginal) != MH_OK)
		throw std::runtime_error("Failed to create hook: " + name);
	if (MH_EnableHook((LPVOID) pTarget) != MH_OK)
		throw std::runtime_error("Failed to enable hook: " + name);
}

void install() {
	SignatureScanner SigScanner;
	SigScanner.GetProcess();
	module mod = SigScanner.GetModule();
	if (!mod.dwBase)
		throw std::runtime_error("Failed to initialize signature scanner");
	if (MH_Initialize() != MH_OK)
		throw std::runtime_error("Failed to initialize Minhook");

	DWORD addrLobbyClientCommands = SigScanner.FindSignature(mod.dwBase, mod.dwSize, "\x55\x8B\xEC\x83\xE4\xF8\x6A\xFF\x68\x00\x00\x00\x00\x64\xA1\x00\x00\x00\x00\x50\x64\x89\x25\x00\x00\x00\x00\x83\xEC\x40\x53\x56\x8B\x75\x08\x8B\x06\x57\x8B\xD9\x68\x00\x00\x00\x00\x50\x89\x5C\x24\x1C\xE8\x00\x00\x00\x00\x83\xC4\x08\x85\xC0", "??????xxx????xx????xxxx????xxxxxxxxxxxxxx????xxxxxx????xxxxx"); //4AAAA0
	addrLobbySendGreentext = (void (__fastcall *)(const char *, void *, void *, int, int)) SigScanner.FindSignature(mod.dwBase, mod.dwSize, "\x6A\xFF\x68\x00\x00\x00\x00\x64\xA1\x00\x00\x00\x00\x50\x64\x89\x25\x00\x00\x00\x00\x51\x56\x57\x8B\xF1\x68\x00\x00\x00\x00\x8D\x4C\x24\x0C\xE8\x00\x00\x00\x00\x80\x7C\x24\x00\x00", "???????xx????xxxx????xxxxxx????xxxxx????xxx??"); //4AA870
	addrMyNickname = *(char **) ((DWORD) addrLobbySendGreentext + 0x4D);
	origLobbyDisplayMessage = (int (__stdcall *)(int, char *)) SigScanner.FindSignature(mod.dwBase, mod.dwSize, "\x55\x8B\xEC\x83\xE4\xF8\x64\xA1\x00\x00\x00\x00\x6A\xFF\x68\x00\x00\x00\x00\x50\x64\x89\x25\x00\x00\x00\x00\x83\xEC\x30\x53\x56\x57\xE8\x00\x00\x00\x00\x33\xC9\x85\xC0\x0F\x95\xC1\x85\xC9\x75\x0A\x68\x00\x00\x00\x00\xE8\x00\x00\x00\x00", "??????xx????xxx????xxxx????xxxxxxx????xxxxxxxxxxxx????x????"); //0x493C50

	DWORD addrConstructResources = SigScanner.FindSignature(mod.dwBase, mod.dwSize,  "\x6A\xFF\x68\x00\x00\x00\x00\x64\xA1\x00\x00\x00\x00\x50\x64\x89\x25\x00\x00\x00\x00\x83\xEC\x24\x53\x55\x8B\x6C\x24\x3C\x8B\x85\x00\x00\x00\x00\x8B\x48\x24", "???????xx????xxxx????xxxxxxxxxxx????xxx"); //5263C0
	DWORD addrOnChatInput = SigScanner.FindSignature(mod.dwBase, mod.dwSize, "\x81\xEC\x00\x00\x00\x00\x55\x56\x57\x8B\xF8\x8A\x07\x84\xC0\x8B\xF1\x0F\x84\x00\x00\x00\x00\x3C\x2F\x0F\x85\x00\x00\x00\x00\x8D\x44\x24\x40", "??????xxxxxxxxxxxxx????xxxx????xxxx"); //0052B5F0

	origShowChatMessage = (int (__stdcall *)(DWORD, int, const char *, int)) SigScanner.FindSignature(mod.dwBase, mod.dwSize, "\x81\xEC\x00\x00\x00\x00\x53\x55\x8B\xAC\x24\x00\x00\x00\x00\x80\xBD\x00\x00\x00\x00\x00\x8B\x85\x00\x00\x00\x00\x8B\x48\x24\x56\x8B\xB1\x00\x00\x00\x00\x57", "??????xxxxx????xx?????xx????xxxxxx????x"); // , 0x52AB70

	hook("ConstructResources", addrConstructResources, (DWORD*)&hookConstructResources, (DWORD*)&origConstructResources);
	hook("OnChatInput", addrOnChatInput, (DWORD *) &hookOnChatInput, (DWORD *) &origOnChatInput);
	hook("LobbyClientCommands", addrLobbyClientCommands, (DWORD *) &hookLobbyClientCommands, (DWORD *) &origLobbyClientCommands);
}

BOOL APIENTRY DllMain(HMODULE hModule, DWORD ul_reason_for_call, LPVOID lpReserved) {
	switch (ul_reason_for_call) {
		case DLL_PROCESS_ATTACH:
			try {
				scanDir(macroDir);
				install();
			} catch (std::exception &e) {
				MessageBoxA(0, e.what(), "wkTextMacros-v1", MB_ICONERROR);
			}
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
		default:
			break;
	}
	return TRUE;
}
